import sqlalchemy
from entity import *

class Schedule:
  def __init__( self, db_filename ):
    self.db_filename = db_filename

    self.engine = sqlalchemy.create_engine('sqlite:///%s'%self.db_filename, echo=False)
    Session = sqlalchemy.orm.sessionmaker(bind=self.engine)
    self.session = Session()

  @property
  def routes(self):
    return self.session.query(Route).all()

  @property
  def agencies(self):
    return self.session.query(Agency).all()

  @property
  def service_periods(self):
    return self.session.query(ServicePeriod).all()

  @property
  def stops(self):
    return self.session.query(Stop).all()

  def create_tables( self ):
    metadata.create_all(self.engine) 



from ..types import make_gtfs_foreign_key_class, Boolean, Time

class Entity(object):
  def __init__(self, **kwargs):
    for attrname, attrtype in self.FIELDS:
      if attrname in kwargs:
        attrvaluestr = kwargs[attrname]
	if attrvaluestr == '':
	  attrvalue = None
	else:
	  attrvalue = attrtype( attrvaluestr )

        setattr( self, attrname, attrvalue )

class Agency(Entity):
  TABLENAME = "agency"
  FIELDS = (('agency_id',str),
            ('agency_name',str),
	    ('agency_url',str),
	    ('agency_timezone',str),
	    ('agency_lang',str),
	    ('agency_phone',str))
  ID_FIELD = "agency_id"

  def __repr__(self):
    return "<Agency %s>"%self.agency_id

  def __init__(self,**kwargs):
    Entity.__init__(self, **kwargs)

    if not hasattr( self, "agency_id" ) or self.agency_id is None:
      self.agency_id = "__DEFAULT__"

class ServicePeriod(Entity):
  TABLENAME = "calendar"
  FIELDS = (('service_id', str),
            ('monday', Boolean),
	    ('tuesday', Boolean),
	    ('wednesday', Boolean),
	    ('thursday', Boolean),
	    ('friday', Boolean),
	    ('saturday', Boolean),
	    ('sunday', Boolean),
	    ('start_date', str),
	    ('end_date',str))
  ID_FIELD = "service_id"

  def __repr__(self):
    return "<ServicePeriod %s %s%s%s%s%s%s%s>"%(self.service_id,
                                           self.monday,
					   self.tuesday,
					   self.wednesday,
					   self.thursday,
					   self.friday,
					   self.saturday,
					   self.sunday)

class ServiceException(Entity):
  TABLENAME = "calendar_dates"
  FIELDS = (('service_id', make_gtfs_foreign_key_class(ServicePeriod)),
            ('date', str),
	    ('exception_type', str))
  ID_FIELD = None

  def __repr__(self):
    return "<ServiceException %s %s>"%(self.date, self.exception_type)

class Route(Entity):
  TABLENAME = "routes"
  FIELDS = (('route_id',str),
            ('agency_id',make_gtfs_foreign_key_class(Agency)),
	    ('route_short_name',str),
	    ('route_long_name',str),
	    ('route_desc',str),
	    ('route_type',int),
	    ('route_url',str),
	    ('route_color',str),
	    ('route_text_color',str))
  ID_FIELD = "route_id"

  def __repr__(self):
    return "<Route %s>"%self.route_id

  def __init__(self, **kwargs):
    Entity.__init__(self, **kwargs)

    if not hasattr( self, "agency_id" ) or self.agency_id is None:
      self.agency_id = "__DEFAULT__"

class Stop(Entity):
  TABLENAME = "stops"
  FIELDS = (('stop_id',str),
            ('stop_code',str),
	    ('stop_name',str),
	    ('stop_desc',str),
	    ('stop_lat',float),
	    ('stop_lon',float),
	    ('zone_id',str),
	    ('stop_url',str),
	    ('location_type',str),
	    ('parent_station',str))
  ID_FIELD = "stop_id"

  def __repr__(self):
    return "<Stop %s>"%self.stop_id

class Trip(Entity):
  TABLENAME = "trips"
  FIELDS = (('route_id',make_gtfs_foreign_key_class(Route)),
            ('service_id',make_gtfs_foreign_key_class(ServicePeriod)),
	    ('trip_id',str),
	    ('trip_headsign',str),
	    ('trip_short_name',str),
	    ('direction_id',str),
	    ('block_id',str),
	    ('shape_id',str))
  ID_FIELD = "trip_id"

  def __repr__(self):
    return "<Trip %s>"%self.trip_id

class StopTime(Entity):
  TABLENAME = "stop_times"
  FIELDS = (('trip_id',make_gtfs_foreign_key_class(Trip)),
            ('arrival_time',Time),
	    ('departure_time',Time),
	    ('stop_id',make_gtfs_foreign_key_class(Stop)),
	    ('stop_sequence',int),
	    ('stop_headsign',str),
	    ('pickup_type',str),
	    ('drop_off_type',str),
	    ('shape_dist_traveled',float))
  ID_FIELD = None
  
  def __repr__(self):
    return "<StopTime %s %s>"%(self.trip_id,self.departure_time)

class Fare(Entity):
  TABLENAME = "fare_attributes"
  FIELDS = (('fare_id',str),
            ('price',str),
	    ('currency_type',str),
	    ('payment_method',str),
	    ('transfers',int),
	    ('transfer_duration',str))
  ID_FIELD = 'fare_id'

  def __repr__(self):
    return "<Fare %s %s>"%(self.price,self.currency_type)

class FareRule(Entity):
  TABLENAME = "fare_rules"
  FIELDS = (('fare_id',make_gtfs_foreign_key_class(Fare)),
            ('route_id',make_gtfs_foreign_key_class(Route)),
	    ('origin_id',str),
	    ('destination_id',str),
	    ('contains_id',str))
  ID_FIELD = None

class ShapePoint(Entity):
  TABLENAME = "shapes"
  FIELDS = (('shape_id',str),
            ('shape_pt_lat',str),
	    ('shape_pt_lon',str),
	    ('shape_pt_sequence',int),
	    ('shape_dist_traveled',str))
  ID_FIELD = None

class Frequency(Entity):
  TABLENAME = "frequencies"
  FIELDS = (('trip_id',make_gtfs_foreign_key_class(Trip)),
            ('start_time',str),
	    ('end_time',str),
	    ('headway_secs',int))
  ID_FIELD = None

  def __repr__(self):
    return "<Frequency %s-%s %s>"%(self.start_time,self.end_time,self.headway_secs)

class Transfer(Entity):
  TABLENAME = "transfers"
  FIELDS = (('from_stop_id',make_gtfs_foreign_key_class(Stop)),
            ('to_stop_id',make_gtfs_foreign_key_class(Stop)),
	    ('transfer_type',int),
	    ('min_transfer_time',str))
  ID_FIELD = None



import os
from codecs import iterdecode
from zipfile import ZipFile
import csv

class Record(object):
  """A Record is a single row in a CSV file"""

  def __init__(self,header,row):
    self.header = header
    self.row = row 

  def to_dict(self):
    return dict([(fieldname,self.row[fieldindex] if fieldindex<len(self.row) else None) for fieldname,fieldindex in self.header.items()])

  def __repr__(self):
    return repr(self.to_dict())

  def __getitem__(self,name):
    try:
      return self.row[ self.header[name] ]
    except KeyError:
      return None

class Table(object):
  """A Table is a single CSV file"""

  def __init__(self, header, rows):

    # header is a dict of name->index
    self.header = dict( zip( header, range(len(header)) ) ) 
    self.rows = rows

  def __repr__(self):
    return "<Table %s>"%self.header

  def __iter__(self):
    return self

  def next(self):
    return Record( self.header, self.rows.next() )

class Feed(object):
  """A Feed is a collection of CSV files with headers, either zipped into an archive
     or loose in a folder"""

  def __init__(self, filename):
    self.filename = filename 
    self.zf = None

    if not os.path.isdir( filename ):
      self.zf = ZipFile( filename )

  def get_rows(self, filename):
    if self.zf:
      try:
        contents = self.zf.read(filename)
      except KeyError:
        raise KeyError( "%s is not present feed"%filename )
      return csv.reader( iterdecode( contents.split("\n"), "utf-8" ) )
    else:
      return csv.reader( iterdecode( open( os.path.join( self.filename, filename ) ), "utf-8" ) )

  def get_table(self, filename):
    rows = self.get_rows( filename )
    return Table( rows.next(), rows )



from ..loader import load
from ..schedule import Schedule

from optparse import OptionParser
import os

def main():
  usage = "usage: %prog [options] gtfs_filename"
  parser = OptionParser(usage)
  parser.add_option( "-o", "--output_filename", dest="output_filename" )

  options, args = parser.parse_args()

  if len(args) != 1:
    parser.error( "No gtfs filename supplied" )

  gtfs_filename = args[0]

  if options.output_filename:
    output_filename = options.output_filename
  else:
    output_filename = os.path.splitext( gtfs_filename )[0]+".db"

  load( gtfs_filename, output_filename )

if __name__=='__main__':
  main()



from schedule import Schedule
import feed
from entity import *
import sys

def load(feed_filename, db_filename=":memory:"):
  schedule = Schedule( db_filename ) 
  schedule.create_tables()
  
  fd = feed.Feed( feed_filename )

  for gtfs_class in (Agency, 
                     Route, 
		     Stop,
		     Trip, 
		     StopTime,
		     ServicePeriod, 
		     ServiceException, 
		     Fare,
		     FareRule,
		     ShapePoint,
		     Frequency,
		     Transfer,
		     ):

    print "loading %s"%gtfs_class
   
    try:
      for i, record in enumerate( fd.get_table( gtfs_class.TABLENAME+".txt" ) ):
        if i%500==0:
	  sys.stdout.write(".")
	  sys.stdout.flush()
	  schedule.session.commit()

        instance = gtfs_class( **record.to_dict() )
        schedule.session.add( instance )
      print
    except KeyError:
      # TODO: check if the table is required
      continue

  schedule.session.commit()

  return schedule



import sqlalchemy
from sqlalchemy.orm import relationship
from unmapped_entities import *
from .. import types

class BooleanType(sqlalchemy.types.TypeDecorator):
  impl = sqlalchemy.types.Integer

  def process_bind_param( self, value, dialect ):
    return int(value)

  def process_result_value( self, value, dialect ):
    return types.Boolean( value )

class TimeType(sqlalchemy.types.TypeDecorator):
  impl = sqlalchemy.types.Integer

  def process_bind_param( self, value, dialect ):
    return value.val if value else None

  def process_result_value( self, value, dialect ):
    return types.Time( value ) if value else None

def table_def_from_entity(entity_class, metadata):
  sqlalchemy_types = {str:sqlalchemy.String,
                      int:sqlalchemy.Integer,
		      float:sqlalchemy.Float,
		      types.Boolean:BooleanType,
		      types.Time:TimeType}
  columns = []
  for field_name,field_type in entity_class.FIELDS:
    if issubclass(field_type, types.ForeignKey):
      foreign_key_column_name = field_type._cls.TABLENAME+"."+field_type._cls.ID_FIELD
      columns.append( sqlalchemy.Column( field_name, sqlalchemy.String, sqlalchemy.ForeignKey(foreign_key_column_name) ) )
    else:
      columns.append( sqlalchemy.Column( field_name,
                              sqlalchemy_types[field_type],
                              primary_key=(field_name==entity_class.ID_FIELD) ) )
  if entity_class.ID_FIELD is None:
    columns.append( sqlalchemy.Column( "id", sqlalchemy.Integer, primary_key=True ) )
  
  return sqlalchemy.Table( entity_class.TABLENAME, metadata, *columns )

def create_and_map_tables(metadata):
  # create the tables
  agency_table = table_def_from_entity( Agency, metadata )
  routes_table = table_def_from_entity( Route, metadata )
  trips_table = table_def_from_entity( Trip, metadata )
  stop_times_table = table_def_from_entity( StopTime, metadata )
  calendar_table = table_def_from_entity( ServicePeriod, metadata )
  calendar_dates_table = table_def_from_entity( ServiceException, metadata )
  fare_attributes_table = table_def_from_entity( Fare, metadata )
  fare_rules_table = table_def_from_entity( FareRule, metadata )
  stops_table = table_def_from_entity( Stop, metadata )
  shapes_table = table_def_from_entity( ShapePoint, metadata )
  frequencies_table = table_def_from_entity( Frequency, metadata )
  transfers_table = table_def_from_entity( Transfer, metadata )

  # map the tables
  sqlalchemy.orm.mapper(Agency, agency_table) 
  sqlalchemy.orm.mapper(Route, routes_table, properties={'agency':relationship(Agency,backref="routes")})
  sqlalchemy.orm.mapper(Stop, stops_table)
  sqlalchemy.orm.mapper(Trip, trips_table, properties={'route':relationship(Route, backref="trips"),
					'service_period':relationship(ServicePeriod, backref="trips")})
  sqlalchemy.orm.mapper(StopTime, stop_times_table, properties={'trip':relationship(Trip, backref="stop_times"),
                                                                'stop':relationship(Stop, backref="stop_times")})
  sqlalchemy.orm.mapper(ServicePeriod, calendar_table)
  sqlalchemy.orm.mapper(ServiceException, calendar_dates_table, properties={'service_period':relationship(ServicePeriod,backref="exceptions")})
  sqlalchemy.orm.mapper(Fare, fare_attributes_table)
  sqlalchemy.orm.mapper(FareRule, fare_rules_table, properties={'fare':relationship(Fare,backref="rules"),
                                                                'route':relationship(Route,backref="fare_rules")})
  sqlalchemy.orm.mapper(ShapePoint, shapes_table)
  sqlalchemy.orm.mapper(Frequency, frequencies_table, properties={'trip':relationship(Trip,backref="frequencies")})
  sqlalchemy.orm.mapper(Transfer, transfers_table, properties={"from_stop":relationship(Stop,primaryjoin=transfers_table.c.from_stop_id==stops_table.c.stop_id,backref="transfers_away"),
                                                "to_stop":relationship(Stop,primaryjoin=transfers_table.c.to_stop_id==stops_table.c.stop_id,backref="transfers_from")})




import re

class ForeignKey(str):
  pass

class Boolean(object):
  def __init__(self, val):
    val = int(val)
    if val not in (0,1):
      raise ValueError( "Boolean values must be '0' or '1'" )
    self.val = val

  def __int__(self):
    return self.val

  def __repr__(self):
    return repr(self.val)


class Time(object):

  def __init__(self, timerepr):
    if isinstance( timerepr, int ):
      self.val = timerepr
    elif isinstance( timerepr, basestring ):
      self.val = self._time_to_seconds_since_midnight( timerepr )
    else:
      raise Exception( "timerepr must be str or int, found %s"%type(timerepr) )
    
  def _time_to_seconds_since_midnight(self, time_string):
    """Convert HHH:MM:SS into seconds since midnight.

    For example "01:02:03" returns 3723. The leading zero of the hours may be
    omitted. HH may be more than 23 if the time is on the following day."""
    m = re.match(r'(\d{1,3}):([0-5]\d):([0-5]\d)$', time_string)
    # ignored: matching for leap seconds
    if not m:
      raise ValueError( 'Bad HH:MM:SS "%s"' % time_string )
    return int(m.group(1)) * 3600 + int(m.group(2)) * 60 + int(m.group(3))
  
  def _format_seconds_since_midnight(self, s):
    """Formats an int number of seconds past midnight into a string
    as "HH:MM:SS"."""
    return "%02d:%02d:%02d" % (s / 3600, (s / 60) % 60, s % 60)

  def __repr__(self):
    return "<Time %s>"%self._format_seconds_since_midnight(self.val)

def make_gtfs_foreign_key_class(cls):
  class ret(ForeignKey):
    _cls = cls
  return ret




from setuptools import setup, find_packages
setup(
    name = "gtfs",
    version = "0.1",
    packages = find_packages(),
    entry_points = {
      "console_scripts": ["compile_gtfs = gtfs.scripts.compile_gtfs:main"]
    }
)
